@@ -1,8 +1,7 @@
-// Copyright (c) 2017-present PyO3 Project and Contributors
-
 use crate::{
     method::{FnArg, FnSpec},
     pyfunction::FunctionSignature,
+    quotes::some_wrap,
 };
 use proc_macro2::{Span, TokenStream};
 use quote::{quote, quote_spanned};
@@ -29,7 +28,6 @@ pub fn is_forwarded_args(signature: &FunctionSignature<'_>) -> bool {
 pub fn impl_arg_params(
     spec: &FnSpec<'_>,
     self_: Option<&syn::Type>,
-    py: &syn::Ident,
     fastcall: bool,
 ) -> Result<(TokenStream, Vec<TokenStream>)> {
     let args_array = syn::Ident::new("output", Span::call_site());
@@ -41,12 +39,12 @@ pub fn impl_arg_params(
             .signature
             .arguments
             .iter()
-            .map(|arg| impl_arg_param(arg, &mut 0, py, &args_array))
+            .map(|arg| impl_arg_param(arg, &mut 0, &args_array))
             .collect::<Result<_>>()?;
         return Ok((
             quote! {
-                let _args = #py.from_borrowed_ptr::<_pyo3::types::PyTuple>(_args);
-                let _kwargs: ::std::option::Option<&_pyo3::types::PyDict> = #py.from_borrowed_ptr_or_opt(_kwargs);
+                let _args = py.from_borrowed_ptr::<_pyo3::types::PyTuple>(_args);
+                let _kwargs: ::std::option::Option<&_pyo3::types::PyDict> = py.from_borrowed_ptr_or_opt(_kwargs);
             },
             arg_convert,
         ));
@@ -79,7 +77,7 @@ pub fn impl_arg_params(
         .signature
         .arguments
         .iter()
-        .map(|arg| impl_arg_param(arg, &mut option_pos, py, &args_array))
+        .map(|arg| impl_arg_param(arg, &mut option_pos, &args_array))
         .collect::<Result<_>>()?;
 
     let args_handler = if spec.signature.python_signature.varargs.is_some() {
@@ -103,7 +101,7 @@ pub fn impl_arg_params(
     let extract_expression = if fastcall {
         quote! {
             DESCRIPTION.extract_arguments_fastcall::<#args_handler, #kwargs_handler>(
-                #py,
+                py,
                 _args,
                 _nargs,
                 _kwnames,
@@ -113,7 +111,7 @@ pub fn impl_arg_params(
     } else {
         quote! {
             DESCRIPTION.extract_arguments_tuple_dict::<#args_handler, #kwargs_handler>(
-                #py,
+                py,
                 _args,
                 _kwargs,
                 &mut #args_array
@@ -145,7 +143,6 @@ pub fn impl_arg_params(
 fn impl_arg_param(
     arg: &FnArg<'_>,
     option_pos: &mut usize,
-    py: &syn::Ident,
     args_array: &syn::Ident,
 ) -> Result<TokenStream> {
     // Use this macro inside this function, to ensure that all code generated here is associated
@@ -155,7 +152,7 @@ fn impl_arg_param(
     }
 
     if arg.py {
-        return Ok(quote_arg_span! { #py });
+        return Ok(quote! { py });
     }
 
     let name = arg.name;
@@ -196,12 +193,7 @@ fn impl_arg_param(
     // Option<T> arguments have special treatment: the default should be specified _without_ the
     // Some() wrapper. Maybe this should be changed in future?!
     if arg.optional.is_some() {
-        default = Some(match &default {
-            Some(expression) if expression.to_string() != "None" => {
-                quote!(::std::option::Option::Some(#expression))
-            }
-            _ => quote!(::std::option::Option::None),
-        })
+        default = Some(default.map_or_else(|| quote!(::std::option::Option::None), some_wrap));
     }
 
     let tokens = if let Some(expr_path) = arg.attrs.from_py_with.as_ref().map(|attr| &attr.value) {
